include "LaunchServices_h.pxi"

from Carbon.File import FSRef as _FSRef
import Enumerators
from util import SimpleStructure
import operator
import _CF
import MacOS

cdef int CheckError(int _err) except -1:
    if (_err == noErr):
        return 0
    if _err in Enumerators.LSErrors:
        raise MacOS.Error(_err, Enumerators.LSErrors[_err])
    else:
        PyMac_Error(_err)
        raise
    return -1
    
cdef object ItemInfoRecordToStructure(LSItemInfoRecord* itemInfo):
    flags = Enumerators.LSItemInfoFlags.fromFlags(itemInfo.flags)
    filetype = PyMac_BuildOSType(itemInfo.filetype)
    creator = PyMac_BuildOSType(itemInfo.creator)
    if itemInfo.extension == NULL:
        extension = None
    else:
        extension = CFObj_New(CFRetain(itemInfo.extension)).toPython()
    if itemInfo.iconFileName == NULL:
        iconFileName = None
    else:
        iconFileName = CFObj_New(CFRetain(itemInfo.iconFileName)).toPython()
    return SimpleStructure('LSItemInfoRecord', 
        flags=flags,
        filetype=filetype,
        creator=creator,
        extension=extension,
        iconFileName=iconFileName,
        kindID=itemInfo.kindID,
    )

def CFRetainForReal(obj):
    cdef CFTypeRef tref
    CFObj_Convert(obj, &tref)
    CFRetain(tref)
    return obj

def CFStringFromObject(obj):
    """
    CFStringFromObject(obj) -> _CF.CFStringRef
    """
    if isinstance(obj, _CF.CFStringRef):
        return obj
    if isinstance(obj, _CF.CFURLRef):
        url = obj.CFURLCopyAbsoluteURL()
        return url.CFURLGetString()
    cfString = _CF.toCF(obj)
    if not isinstance(cfString, _CF.CFStringRef):
        raise ValueError, 'Must be a string'
    return cfString

def CFURLFromObject(obj, base=None):
    """
    CFURLFromObject(obj, base=None) -> _CF.CFURLRef
    """
    cdef CFStringRef _cfString
    cdef CFURLRef _cfURL
    cdef CFURLRef _cfURLbase
    if isinstance(obj, _CF.CFURLRef):
        return obj
    cfString = CFStringFromObject(obj)
    CFStringRefObj_Convert(cfString, &_cfString)
    if base is None:
        _cfURLbase = NULL
    else:
        CFURLRefObj_Convert(CFURLFromObject(base), &_cfURLbase)
    _cfURL = CFURLCreateWithString(kCFAllocatorDefault, _cfString, _cfURLbase)
    if (_cfURL == NULL):
        raise ValueError, 'Not a valid URL'
    return CFObj_New(CFRetain(_cfURL))
    
def PathToURL(path):
    """
    PathToURL(path) -> unicode_url
    """
    path = CFStringFromObject(path).toPython().encode('utf8')
    cfurl = CFRetainForReal(_CF.CFURLCreateFromFSRef(path))
    cfstr = CFStringFromObject(cfurl)
    return cfstr.toPython()

def URLToPath(url):
    """
    URLToPath(url) -> unicode_path
    """
    return CFURLFromObject(url).CFURLCopyFileSystemPath(0).toPython()

def GetInfoForPath(path, *flags):
    """
    GetInfoForPath(path, LSRequestedInfo *flags) -> LSItemInfoRecord

    class LSItemInfoRecord:
        LSItemInfoFlags flags
        OSType filetype
        OSType creator
        unicode extension    (or None)
        unicode iconFileName (or None)
        int LSKindID

    LSCopyItemInfoForRef()
    """
    cdef LSItemInfoRecord itemInfo
    cdef FSRef* fsref
    cdef int fsref_len
    cdef OSErr _err
    cdef int inWhichInfo 
    inWhichInfo = reduce(operator.__or__, flags, 0)
    
    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&fsref, &fsref_len)
    
    _err = LSCopyItemInfoForRef(fsref, inWhichInfo, &itemInfo)
    CheckError(_err)

    return ItemInfoRecordToStructure(&itemInfo)

def GetInfoForURL(url, *flags):
    """
    GetInfoForURL(url, LSRequestedInfo *flags) -> LSItemInfoRecord

    class LSItemInfoRecord:
        LSItemInfoFlags flags
        OSType filetype
        OSType creator
        unicode extension    (or None)
        unicode iconFileName (or None)
        int LSKindID

    LSCopyItemInfoForURL()
    """
    cdef LSItemInfoRecord itemInfo
    cdef CFURLRef _inItemURL
    cdef OSErr _err
    cdef int inWhichInfo 
    inWhichInfo = reduce(operator.__or__, flags, 0)
    cf_inItemURL = CFURLFromObject(url)
    CFURLRefObj_Convert(cf_inItemURL, &_inItemURL)
    
    _err = LSCopyItemInfoForURL(_inItemURL, inWhichInfo, &itemInfo)
    CheckError(_err)

    return ItemInfoRecordToStructure(&itemInfo)

def GetExtensionInfoForPath(path):
    """
    GetExtensionInfoForPath(path) -> (unicode_filename, unicode_extension)

    LSGetExtensionInfo()
    """
    uniname = unicode(path)
    cdef UniCharCount outidx
    cdef OSErr _err

    _err = LSGetExtensionInfo(len(uniname), PyUnicode_AS_UNICODE(uniname), &outidx)
    CheckError(_err)

    if outidx == kLSInvalidExtensionIndex:
        return (uniname, unicode())
    return (uniname[:outidx-1], uniname[outidx-1:])

def GetDisplayNameForPath(path):
    """
    GetDisplayNameForPath(path) -> unicode_displayname

    Takes into consideration whether this item has a hidden extension or not.

    LSCopyDisplayNameForRef()
    """
    cdef FSRef* fsref
    cdef int fsref_len
    cdef OSErr _err
    cdef CFStringRef displayName
    
    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&fsref, &fsref_len)
     
    _err = LSCopyDisplayNameForRef(fsref, &displayName)
    CheckError(_err)

    return CFObj_New(CFRetain(displayName)).toPython()

def GetDisplayNameForURL(url):
    """
    GetDisplayNameForURL(url) -> unicode_displayname

    Takes into consideration whether this item has a hidden extension or not.

    LSCopyDisplayNameForURL()
    """
    cdef OSErr _err
    cdef CFStringRef displayName
    cdef CFURLRef _inURL
    
    CFURLRefObj_Convert(CFURLFromObject(url), &_inURL)
    _err = LSCopyDisplayNameForURL(_inURL, &displayName)
    CheckError(_err)

    return CFObj_New(CFRetain(displayName)).toPython()



def SetExtensionHiddenForPath(path, hide=True):
    """
    SetExtensionHidden(path, hide=True) -> None

    Sets the necessary file system state to indicate that the
    extension for path is hidden, as in the Finder. You can
    determine if a path's extension is hidden using
    GetItemInfoForPath.

    LSSetExtensionHiddenForRef()
    """
    cdef FSRef* fsref
    cdef int fsref_len
    cdef OSErr _err
    cdef Boolean doHide
    
    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&fsref, &fsref_len)
     
    if hide:
        doHide = 1
    else:
        doHide = 0
    
    _err = LSSetExtensionHiddenForRef(fsref, doHide)
    CheckError(_err)

def SetExtensionHiddenForURL(url, hide=True):
    """
    SetExtensionHiddenForURL(url, hide=True) -> None

    Sets the necessary file system state to indicate that the
    extension for path is hidden, as in the Finder. You can
    determine if a URL's extension is hidden using
    GetItemInfoForURL.

    LSSetExtensionHiddenForURL()
    """

    cdef OSErr _err
    cdef Boolean doHide
    cdef CFURLRef _inURL
    
    CFURLRefObj_Convert(CFURLFromObject(url), &_inURL)
    if hide:
        doHide = 1
    else:
        doHide = 0
    
    _err = LSSetExtensionHiddenForURL(_inURL, doHide)
    CheckError(_err)

def GetKindStringForPath(path):
    """
    GetKindStringForPath(path) -> unicode_kind
    
    Returns the kind string as used in the Finder and elsewhere. 

    LSCopyKindStringForRef()
    """
    cdef FSRef* fsref
    cdef int fsref_len
    cdef OSErr _err
    cdef CFStringRef displayName

    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&fsref, &fsref_len)
     
    _err = LSCopyKindStringForRef(fsref, &displayName)
    CheckError(_err)

    return CFObj_New(CFRetain(displayName)).toPython()

def GetKindStringForURL(object url):
    """
    GetKindStringForPath(url) -> unicode_kind
    
    Returns the kind string as used in the Finder and elsewhere. 

    LSCopyKindStringForURL()
    """
    cdef OSErr _err
    cdef CFStringRef displayName

    cdef CFURLRef _inURL
    
    CFURLRefObj_Convert(CFURLFromObject(url), &_inURL)
    _err = LSCopyKindStringForURL(_inURL, &displayName)
    CheckError(_err)

    return CFObj_New(CFRetain(displayName)).toPython()

    
def GetKindStringForTypeInfo(filetype=None, creator=None, extension=None):
    """
    GetKindStringForTypeInfo(
        filetype    =   None,   -- OSType file type code
        creator     =   None,   -- OSType creator code
        extension   =   None,   -- File extension (without '.')
    ) -> unicode_kind

    LSCopyKindStringForTypeInfo()
    """
    cdef OSType _inType
    cdef OSType _inCreator
    cdef CFStringRef _inExtension
    cdef CFStringRef _outKindString
    cdef OSErr _err
    _inType = kLSUnknownType
    _inCreator = kLSUnknownCreator
    _inExtension = NULL
    if filetype:
        PyMac_GetOSType(filetype, &_inType)
    if creator:
        PyMac_GetOSType(creator, &_inCreator)
    if extension:
        cf_inExtension = CFStringFromObject(extension)
        CFStringRefObj_Convert(cf_inExtension, &_inExtension)

    _err = LSCopyKindStringForTypeInfo(_inType, _inCreator, _inExtension, &_outKindString)
    CheckError(_err)
    
    return CFObj_New(CFRetain(_outKindString)).toPython()

def GetKindStringForMIMEType(mimetype):
    """
    GetKindStringForTypeInfo(mimetype) -> unicode_kind

    LSCopyKindStringForMIMEType()
    """
    cdef CFStringRef _inMimeType
    cdef CFStringRef _outKindString
    cf_inMimeType = CFStringFromObject(mimetype)
    CFStringRefObj_Convert(cf_inMimeType, &_inMimeType)
    
    _err = LSCopyKindStringForMIMEType(_inMimeType, &_outKindString)
    CheckError(_err)
    
    return CFObj_New(CFRetain(_outKindString)).toPython()

def GetApplicationPathForPath(path, *flags):
    """
    GetApplicationPathForPath(
        path,               -- path for item
        LSRolesMask* flags  -- defaults to kLSRolesAll
    ) -> unicode_path
    
    Consults the binding tables to return the application that would
    be used to open path if it were double-clicked in the
    Finder. This application will be the user-specified override if
    appropriate or the default otherwise. If no application is known
    to LaunchServices suitable for opening this item,
    kLSApplicationNotFoundErr will be raised.

    LSGetApplicationForItem()
    """

    cdef FSRef* fsref
    cdef FSRef out_fsref
    cdef int fsref_len
    cdef OSErr _err
    cdef int inRoleMask

    if flags == ():
        inRoleMask = kLSRolesAll
    else:
        inRoleMask = reduce(operator.__or__, flags, 0)
    
    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&fsref, &fsref_len)

    _err = LSGetApplicationForItem(fsref, inRoleMask, &out_fsref, NULL)
    CheckError(_err)
    
    return _FSRef(rawdata=PyString_FromStringAndSize(<char *>&out_fsref, sizeof(FSRef))).as_pathname().decode('utf8')
    
def GetApplicationPathForInfo(filetype=None, creator=None, extension=None, *flags):
    """
    GetApplicationPathForInfo(
        filetype    =   None,   -- OSType file type code
        creator     =   None,   -- OSType creator code
        extension   =   None,   -- File extension (without '.')
        LSRolesMask* flags      -- defaults to kLSRolesAll
    ) -> unicode_path
    
    Consults the binding tables to return the application that would
    be used to open items with type, creator, and/or extension as
    provided if they were double-clicked in the Finder. This
    application will be the default for items like this if one has
    been set. If no application is known to LaunchServices suitable
    for opening such items, kLSApplicationNotFoundErr will be
    raised. Not all three input parameters can be None at the same
    time.

    LSGetApplicationForInfo()
    """
    cdef OSType _inType
    cdef OSType _inCreator
    cdef CFStringRef _inExtension
    cdef FSRef out_fsref
    cdef int inRoleMask
    cdef OSErr _err
    _inType = kLSUnknownType
    _inCreator = kLSUnknownCreator
    _inExtension = NULL
    if flags == ():
        inRoleMask = kLSRolesAll
    else:
        inRoleMask = reduce(operator.__or__, flags, 0)
    if filetype:
        PyMac_GetOSType(filetype, &_inType)
    if creator:
        PyMac_GetOSType(creator, &_inCreator)
    if extension:
        cf_inExtension = CFStringFromObject(extension)
        CFStringRefObj_Convert(cf_inExtension, &_inExtension)

    _err = LSGetApplicationForInfo(_inType, _inCreator, _inExtension, inRoleMask, &out_fsref, NULL)
    CheckError(_err)
    
    return _FSRef(rawdata=PyString_FromStringAndSize(<char *>&out_fsref, sizeof(FSRef))).as_pathname().decode('utf8')

def GetApplicationURLForMIMEType(mimetype, *flags):
    """
    GetApplicationURLForMIMEType(
        mimetype,                   -- mime/type
        LSRolesMask* flags,         -- defaults to kLSRolesAll
    ) -> unicode_url

    The returned application URL will be the user's preferred handler
    for the MIME type if one has been set. If no user preferred
    application has been set, Launch Services will select a default
    handler for the MIME type. If no application is known to handle
    the MIME type, kLSApplicationNotFoundErr will be raised.

    LSCopyApplicationForMIMEType()
    """
    cdef CFStringRef _inMIMEType
    cdef int inRoleMask
    cdef CFURLRef outAppURL
    
    cf_inMIMEType = CFStringFromObject(mimetype)
    CFStringRefObj_Convert(cf_inMIMEType, &_inMIMEType)
    
    if flags == ():
        inRoleMask = kLSRolesAll
    else:
        inRoleMask = reduce(operator.__or__, flags, 0)
    
    _err = LSCopyApplicationForMIMEType(_inMIMEType, inRoleMask, &outAppURL)
    CheckError(_err)
        
    return CFStringFromObject(CFObj_New(CFRetain(outAppURL))).toPython()

def GetApplicationURLForURL(url, *flags):
    """
    GetApplicationURLForURL(
        url,                    -- url to query
        LSRolesMask* flags      -- defaults to kLSRolesAll
    ) -> unicode_url

    Consults the binding tables to return the application that would
    be used to open url if it were double-clicked in the Finder.
    This application will be the user-specified override if
    appropriate or the default otherwise. If no application is known
    to LaunchServices suitable for opening this item,
    kLSApplicationNotFoundErr will be raised.

    LSGetApplicationForURL()
    """
    cdef CFURLRef _inURL
    cdef int inRoleMask
    cdef CFURLRef outAppURL
    
    cf_inURL = CFURLFromObject(url)
    CFURLRefObj_Convert(cf_inURL, &_inURL)
    
    if flags == ():
        inRoleMask = kLSRolesAll
    else:
        inRoleMask = reduce(operator.__or__, flags, 0)
    
    _err = LSGetApplicationForURL(_inURL, inRoleMask, NULL, &outAppURL)
    CheckError(_err)
        
    return CFStringFromObject(CFObj_New(CFRetain(outAppURL))).toPython()

def FindApplicationPath(creator=None, bundle=None, name=None):
    """
    FindApplicationPath(
        creator = None,     -- OSType file creator to consider
        bundle  = None,     -- bundle identifier to consider
        name    = None,     -- application name to consider, must include extension!
    ) -> unicode_path

    Returns the application path with the corresponding input information.
    The registry of applications is consulted first in order of
    bundleID, then creator, then name. All comparisons are case
    insensitive and 'ties' are decided first by version, then by
    native vs. Classic.

    LSFindApplicationForInfo()
    """
    cdef OSType _inCreator
    cdef CFStringRef _inBundleID
    cdef CFStringRef _inName
    cdef FSRef out_fsref
    cdef OSErr _err
    _inCreator = kLSUnknownCreator
    _inName = NULL
    _inBundleID = NULL
    if creator:
        PyMac_GetOSType(creator, &_inCreator)
    if name:
        cf_inName = CFStringFromObject(name)
        CFStringRefObj_Convert(cf_inName, &_inName)
    if bundle:
        cf_inBundleID = CFStringFromObject(bundle)
        CFStringRefObj_Convert(cf_inBundleID, &_inBundleID)

    _err = LSFindApplicationForInfo(_inCreator, _inBundleID, _inName, &out_fsref, NULL)
    CheckError(_err)
        
    return _FSRef(rawdata=PyString_FromStringAndSize(<char *>&out_fsref, sizeof(FSRef))).as_pathname().decode('utf8')

def FindApplicationURL(creator=None, bundle=None, name=None):
    """
    FindApplicationURL(
        creator = None,     -- OSType file creator to consider
        bundle  = None,     -- bundle identifier to consider
        name    = None,     -- application name to consider, must include extension!
    ) -> unicode_url
    
    Returns the application url with the corresponding input information.
    The registry of applications is consulted first in order of
    bundleID, then creator, then name. All comparisons are case
    insensitive and 'ties' are decided first by version, then by
    native vs. Classic.

    LSFindApplicationForInfo()
    """
    cdef OSType _inCreator
    cdef CFStringRef _inBundleID
    cdef CFStringRef _inName
    cdef OSErr _err
    cdef CFURLRef outAppURL
    _inCreator = kLSUnknownCreator
    _inName = NULL
    _inBundleID = NULL
    if creator:
        PyMac_GetOSType(creator, &_inCreator)
    if name:
        cf_inName = CFStringFromObject(name)
        CFStringRefObj_Convert(cf_inName, &_inName)
    if bundle:
        cf_inBundleID = CFStringFromObject(bundle)
        CFStringRefObj_Convert(cf_inBundleID, &_inBundleID)

    _err = LSFindApplicationForInfo(_inCreator, _inBundleID, _inName, NULL, &outAppURL)
    CheckError(_err)
        
    return CFStringFromObject(CFObj_New(CFRetain(outAppURL))).toPython()

def CanApplicationPathAcceptPath(app, path, *flags):
    """
    CanApplicationPathAcceptPath(
        app,                -- application path
        path,               -- item path
        LSRolesMask* flags, -- defaults to kLSRolesAll
    ) -> bool

    Returns whether app can accept path as in a drag and drop 
    operation. If flags are other than kLSRolesAll then make 
    sure app claims to fulfill the requested role.

    WARNING:  
    I don't trust the implementation of this.
    I'm pretty sure it's not my fault.  Use the URL version,
    it seems to work (it passes the unit tests).
    
    LSCanRefAcceptItem()
    """
    cdef FSRef* _inItemFSRef
    cdef FSRef* _inTargetRef
    cdef int fsref_len
    cdef OSErr _err
    cdef int inRoleMask
    cdef Boolean res
    if flags == ():
        inRoleMask = kLSRolesAll
    else:
        inRoleMask = reduce(operator.__or__, flags, 0)
    
    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&_inItemFSRef, &fsref_len)
    PyObject_AsReadBuffer(_FSRef(app).data, <char**>&_inTargetRef, &fsref_len)
    _err = LSCanRefAcceptItem(_inItemFSRef, _inTargetRef, inRoleMask, kLSAcceptDefault, &res)
    CheckError(_err)
    
    if res:
        return True 
    else:
        return False

def CanApplicationURLAcceptURL(app, url, *flags):
    """
    CanApplicationURLAcceptURL(
        app,                -- application url
        url,                -- item url
        LSRolesMask* flags, -- defaults to kLSRolesAll
    ) -> bool

    Returns whether app can accept url as in a drag and drop 
    operation. If flags are other than kLSRolesAll then make 
    sure app claims to fulfill the requested role.

    LSCanURLAcceptURL()
    """
    cdef CFURLRef _inItemURL
    cdef CFURLRef _inTargetURL
    cdef int fsref_len
    cdef OSErr _err
    cdef int inRoleMask
    cdef Boolean res
    if flags == ():
        inRoleMask = kLSRolesAll
    else:
        inRoleMask = reduce(operator.__or__, flags, 0)
    
    cf_inItemURL = CFURLFromObject(url)
    CFURLRefObj_Convert(cf_inItemURL, &_inItemURL)
    cf_inTargetURL = CFURLFromObject(app)
    CFURLRefObj_Convert(cf_inTargetURL, &_inTargetURL)

    _err = LSCanURLAcceptURL(_inItemURL, _inTargetURL, inRoleMask, kLSAcceptDefault, &res)
    CheckError(_err)
    
    if res:
        return True
    else:
        return False

def OpenPath(path):
    """
    OpenPath(path) -> unicode_path
    
    Opens applications, documents, and folders. Applications are
    opened via an 'oapp' or 'rapp' event. Documents are opened in
    their user-overridden or default applications as appropriate.
    Folders are opened in the Finder. 

    (TODO)
    Use the more specific OpenPathFromSpec for more control over 
    launching.

    LSOpenFSRef()
    """
    cdef FSRef* fsref
    cdef FSRef out_fsref
    cdef int fsref_len
    cdef OSErr _err
    
    PyObject_AsReadBuffer(_FSRef(path).data, <char**>&fsref, &fsref_len)

    _err = LSOpenFSRef(fsref, &out_fsref)
    CheckError(_err)

    return _FSRef(rawdata=PyString_FromStringAndSize(<char *>&out_fsref, sizeof(FSRef))).as_pathname().decode('utf8')

def OpenURL(url):
    """
    OpenURL(url) -> unicode_url

    Opens applications, documents, and folders. Applications are
    opened via an 'oapp' or 'rapp' event. Documents are opened in
    their user-overridden or default applications as appropriate.
    Folders are opened in the Finder. 

    (TODO)
    Use the more specific OpenURLFromSpec for more control over 
    launching.

    LSOpenCFURLRef()
    """
    cdef CFURLRef _inURL
    cdef CFURLRef outAppURL
    
    cf_inURL = CFURLFromObject(url)
    CFURLRefObj_Convert(cf_inURL, &_inURL)
    
    _err = LSOpenCFURLRef(_inURL, &outAppURL)
    CheckError(_err)
        
    return CFStringFromObject(CFObj_New(CFRetain(outAppURL))).toPython()

def OpenRefFromSpec():
    """
    (TODO)

    LSOpenFromRefSpec()
    """
    raise NotImplementedError

def OpenURLFromSpec():
    """
    (TODO)

    LSOpenFromURLSpec()
    """
    raise NotImplementedError
